@@ -57,6 +57,7 @@ static void         spell_error_destroy     (SpellError *se);
 static void         move_spell_errors       (GList *spell_errors, guint offset, gint delta);
 static GList *      remove_spell_errors     (GList *spell_errors, guint offset, guint len);
 static void         remove_text_slaves      (HTMLObject *self);
+static void         html_text_change_attrs  (PangoAttrList *attr_list, GtkHTMLFontStyle style, HTMLEngine *e, gint start_index, gint end_index);
 
 /* void
 debug_spell_errors (GList *se)
@@ -148,8 +149,8 @@ copy (HTMLObject *s,
 	dest->font_style    = src->font_style;
 	dest->face          = g_strdup (src->face);
 	dest->color         = src->color;
-	dest->select_start  = src->select_start;
-	dest->select_length = src->select_length;
+	dest->select_start  = 0;
+	dest->select_length = 0;
 	dest->attr_list     = pango_attr_list_copy (src->attr_list);
 	dest->extra_attr_list = src->extra_attr_list ? pango_attr_list_copy (src->extra_attr_list) : NULL;
 
@@ -256,7 +257,7 @@ cut_attr_list (HTMLText *text, gint begin_index, gint end_index)
 }
 
 static void
-cut_links (HTMLText *text, gint start_offset, gint end_offset)
+cut_links_full (HTMLText *text, int start_offset, int end_offset, int start_index, int end_index, int shift_offset, int shift_index)
 {
 	GSList *l, *next;
 	Link *link;
@@ -268,13 +269,52 @@ cut_links (HTMLText *text, gint start_offset, gint end_offset)
 		if (start_offset <= link->start_offset && link->end_offset <= end_offset) {
 			html_link_free (link);
 			text->links = g_slist_delete_link (text->links, l);
-		} else if (start_offset <= link->start_offset && link->start_offset <= end_offset)
-			link->start_offset = end_offset;
-		else if (start_offset <= link->end_offset && link->end_offset <= end_offset)
+		} else if (end_offset <= link->start_offset) {
+			link->start_offset -= shift_offset;
+			link->start_index -= shift_index;
+			link->end_offset -= shift_offset;
+			link->end_index -= shift_index;
+		} else if (start_offset <= link->start_offset)  {
+			link->start_offset = end_offset - shift_offset;
+			link->end_offset -= shift_offset;
+			link->start_index = end_index - shift_index;
+			link->end_index -= shift_index;
+		} else if (end_offset <= link->end_offset) {
+			if (shift_offset > 0) {
+				link->end_offset -= shift_offset;
+				link->end_index -= shift_index;
+			} else {
+				if (link->end_offset == end_offset) {
+					link->end_offset = start_offset;
+					link->end_index = start_index;
+				} else if (link->start_offset == start_offset) {
+					link->start_offset = end_offset;
+					link->start_index = end_index;
+				} else {
+					Link *dup = html_link_dup (link);
+
+					link->start_offset = end_offset;
+					link->start_index = end_index;
+					dup->end_offset = start_offset;
+					dup->end_index = start_index;
+
+					l = g_slist_prepend (l, dup);
+					next = l->next->next;
+				}
+			}
+		} else if (start_offset < link->end_offset) {
 			link->end_offset = start_offset;
+			link->end_index = start_index;
+		}
 	}
 }
 
+static void
+cut_links (HTMLText *text, int start_offset, int end_offset, int start_index, int end_index)
+{
+	cut_links_full (text, start_offset, end_offset, start_index, end_index, end_offset - start_offset, end_index - start_index);
+}
+
 HTMLObject *
 html_text_op_copy_helper (HTMLText *text, GList *from, GList *to, guint *len)
 {
@@ -294,21 +334,20 @@ html_text_op_copy_helper (HTMLText *text, GList *from, GList *to, guint *len)
 
 	rv = html_object_dup (HTML_OBJECT (text));
 	rvt = HTML_TEXT (rv);
-
-	if (end_index < rvt->text_bytes)
-		cut_attr_list (rvt, end_index, rvt->text_bytes);
-	if (begin_index > 0)
-		cut_attr_list (rvt, 0, begin_index);
-	if (end < rvt->text_len)
-		cut_links (rvt, end, rvt->text_len);
-	if (begin > 0)
-		cut_links (rvt, 0, begin);
-
-	nt = g_strndup (rvt->text + begin_index, end_index - begin_index);
-	g_free (rvt->text);
-	rvt->text = nt;
 	rvt->text_len = end - begin;
 	rvt->text_bytes = end_index - begin_index;
+	nt = g_strndup (rvt->text + begin_index, rvt->text_bytes);
+	g_free (rvt->text);
+	rvt->text = nt;
+
+	if (end_index < text->text_bytes)
+		cut_attr_list (rvt, end_index, text->text_bytes);
+	if (begin_index > 0)
+		cut_attr_list (rvt, 0, begin_index);
+	if (end < text->text_len)
+		cut_links (rvt, end, text->text_len, end_index, text->text_bytes);
+	if (begin > 0)
+		cut_links (rvt, 0, begin, 0, begin_index);
 
 	return rv;
 }
@@ -350,11 +389,11 @@ html_text_op_cut_helper (HTMLText *text, HTMLEngine *e, GList *from, GList *to,
 			cut_attr_list (rvt, end_index, rvt->text_bytes);
 		if (begin_index > 0)
 			cut_attr_list (rvt, 0, begin_index);
-		cut_links (text, begin, end);
+		cut_links (text, begin, end, begin_index, end_index);
 		if (end < rvt->text_len)
-			cut_links (rvt, end, rvt->text_len);
+			cut_links (rvt, end, rvt->text_len, end_index, rvt->text_bytes);
 		if (begin > 0)
-			cut_links (rvt, 0, begin);
+			cut_links (rvt, 0, begin, 0, begin_index);
 		nt = g_strconcat (text->text, tail, NULL);
 		g_free (text->text);
 
@@ -376,6 +415,7 @@ html_text_op_cut_helper (HTMLText *text, HTMLEngine *e, GList *from, GList *to,
 		move_spell_errors (text->spell_errors, end, - (end - begin));
 
 		html_text_convert_nbsp (text, TRUE);
+		html_text_convert_nbsp (rvt, TRUE);
 		pango_info_destroy (text);
 	} else {
 		text->spell_errors = remove_spell_errors (text->spell_errors, 0, text->text_len);
@@ -420,7 +460,9 @@ merge_links (HTMLText *t1, HTMLText *t2)
 			Link *link = (Link *) l->data;
 
 			link->start_offset += t1->text_len;
+			link->start_index += t1->text_bytes;
 			link->end_offset += t1->text_len;
+			link->end_index += t1->text_bytes;
 		}
 
 		if (t1->links) {
@@ -429,6 +471,7 @@ merge_links (HTMLText *t1, HTMLText *t2)
 
 			if (tail->start_offset == head->end_offset && html_link_equal (head, tail)) {
 				tail->start_offset = head->start_offset;
+				tail->start_index = head->start_index;
 				html_link_free (head);
 				t1->links = g_slist_delete_link (t1->links, t1->links);
 			}
@@ -548,7 +591,7 @@ split_attrs (HTMLText *t1, HTMLText *t2, gint index)
 }
 
 static void
-split_links (HTMLText *t1, HTMLText *t2, gint offset)
+split_links (HTMLText *t1, HTMLText *t2, gint offset, gint index)
 {
 	GSList *l, *prev = NULL;
 
@@ -556,8 +599,10 @@ split_links (HTMLText *t1, HTMLText *t2, gint offset)
 		Link *link = (Link *) l->data;
 
 		if (link->start_offset < offset) {
-			if (link->end_offset > offset)
+			if (link->end_offset > offset) {
 				link->end_offset = offset;
+				link->end_index = index;
+			}
 
 			if (prev) {
 				prev->next = NULL;
@@ -582,6 +627,7 @@ split_links (HTMLText *t1, HTMLText *t2, gint offset)
 		if (link->start_offset < offset) {
 			if (link->end_offset > offset) {
 				link->start_offset = offset;
+				link->start_index = index;
 				prev = l;
 				l = l->next;
 			}
@@ -601,7 +647,9 @@ split_links (HTMLText *t1, HTMLText *t2, gint offset)
 		Link *link = (Link *) l->data;
 
 		link->start_offset -= offset;
+		link->start_index -= index;
 		link->end_offset -= offset;
+		link->end_index -= index;
 	}
 }
 
@@ -633,7 +681,7 @@ object_split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, g
 	t2->text_len   -= offset;
 	t2->text_bytes -= split_index;
 	split_attrs (t1, t2, split_index);
-	split_links (t1, t2, offset);
+	split_links (t1, t2, offset, split_index);
 	if (!html_text_convert_nbsp (t2, FALSE))
 		t2->text = g_strdup (t2->text);
 	g_free (tt);
@@ -678,7 +726,7 @@ object_split (HTMLObject *self, HTMLEngine *e, HTMLObject *child, gint offset, g
 }
 
 static gboolean
-calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
+html_text_real_calc_size (HTMLObject *self, HTMLPainter *painter, GList **changed_objs)
 {
 	self->width = 0;
 	html_object_calc_preferred_width (self, painter);
@@ -1077,16 +1125,14 @@ html_text_get_pango_info (HTMLText *text, HTMLPainter *painter)
 		PangoAttrList *attrs;
 		PangoAttribute *attr;
 		gchar *translated, *heap = NULL;
-		guint bytes;
 		gint i;
 
-		bytes = strlen (text->text);
-		if (bytes > HTML_ALLOCA_MAX)
-			heap = translated = g_malloc (bytes);
+		if (text->text_bytes > HTML_ALLOCA_MAX)
+			heap = translated = g_malloc (text->text_bytes);
 		else 
-			translated = alloca (bytes);
+			translated = alloca (text->text_bytes);
 
-		html_replace_tabs (text->text, translated, bytes);
+		html_replace_tabs (text->text, translated, text->text_bytes);
 		if (HTML_IS_PLAIN_PAINTER (painter)) {
 			attrs = pango_attr_list_new ();
 			attr = pango_attr_family_new (painter->font_manager.fixed.face);
@@ -1103,7 +1149,35 @@ html_text_get_pango_info (HTMLText *text, HTMLPainter *painter)
 			attrs = pango_attr_list_copy (text->attr_list);
 		if (text->extra_attr_list)
 			pango_attr_list_splice (attrs, text->extra_attr_list, 0, text->text_len);
+		if (HTML_OBJECT (text)->parent && HTML_IS_CLUEFLOW (HTML_OBJECT (text)->parent) && painter->widget && GTK_IS_HTML (painter->widget)) {
+			GtkHTMLFontStyle parent_style = html_clueflow_get_default_font_style (HTML_CLUEFLOW (HTML_OBJECT (text)->parent));
 
+			html_text_change_attrs (attrs, parent_style, GTK_HTML (painter->widget)->engine, 0, text->text_bytes);
+		}
+		if (text->links && painter->widget && GTK_IS_HTML (painter->widget)) {
+			PangoAttrList *links_attrs = pango_attr_list_new ();
+			HTMLColor *link_color = html_colorset_get_color (GTK_HTML (painter->widget)->engine->settings->color_set, HTMLLinkColor);
+			GSList *l;
+
+			for (l = text->links; l; l = l->next) {
+				PangoAttribute *attr;
+				Link *link;
+
+				link = (Link *) l->data;
+				attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
+				attr->start_index = link->start_index;
+				attr->end_index = link->end_index;
+				pango_attr_list_insert_before (links_attrs, attr);
+
+				attr = pango_attr_foreground_new (link_color->color.red, link_color->color.green, link_color->color.blue);
+				attr->start_index = link->start_index;
+				attr->end_index = link->end_index;
+				pango_attr_list_change (links_attrs, attr);
+
+			}
+			pango_attr_list_splice (attrs, links_attrs, 0, text->text_len);
+			pango_attr_list_unref (links_attrs);
+		}
 		if (text->select_length) {
 			gchar *end;
 			gchar *start;
@@ -1126,7 +1200,7 @@ html_text_get_pango_info (HTMLText *text, HTMLPainter *painter)
 			pango_attr_list_change (attrs, attr);
 		}
 
-		items = pango_itemize (pc, translated, 0, bytes, attrs, NULL);
+		items = pango_itemize (pc, translated, 0, text->text_bytes, attrs, NULL);
 		pango_attr_list_unref (attrs);
 
 		text->pi = html_text_pango_info_new (g_list_length (items));
@@ -1239,6 +1313,13 @@ update_mw (HTMLText *text, HTMLPainter *painter, gint offset, gint *last_offset,
 	*last_offset = offset;
 }
 
+gboolean
+html_text_is_line_break (PangoItem *item, PangoLogAttr *attrs, int offset)
+{
+	return (attrs [offset].is_line_break
+		&& (item->analysis.lang_engine != NULL || (offset > 0 && attrs [offset - 1].is_white)));
+}
+
 static gint
 calc_min_width (HTMLObject *self, HTMLPainter *painter)
 {
@@ -1257,7 +1338,7 @@ calc_min_width (HTMLObject *self, HTMLPainter *painter)
 	while (offset < text->text_len) {
 		gint skip;
 
-		if (offset > 0 && pi->entries [ii].attrs [io].is_line_break)
+		if (offset > 0 && html_text_is_line_break (pi->entries [ii].item, pi->entries [ii].attrs, io))
 			update_mw (text, painter, offset, &last_offset, &ww, &mw, ii, io, s, line_offset);
 
 		if (*s == '\t') {
@@ -1332,7 +1413,7 @@ save_open_attrs (HTMLEngineSaveState *state, GSList *attrs)
 		case PANGO_ATTR_SIZE:
 			if (attr->klass == &html_pango_attr_font_size_klass) {
 				HTMLPangoAttrFontSize *size = (HTMLPangoAttrFontSize *) attr;
-				if ((size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != GTK_HTML_FONT_STYLE_SIZE_3) {
+				if ((size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != GTK_HTML_FONT_STYLE_SIZE_3 && (size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != 0) {
 					tag = g_strdup_printf ("<FONT SIZE=\"%d\">", size->style & GTK_HTML_FONT_STYLE_SIZE_MASK);
 					free_tag = TRUE;
 				}
@@ -1384,6 +1465,12 @@ save_close_attrs (HTMLEngineSaveState *state, GSList *attrs)
 			tag = "</S>";
 			break;
 		case PANGO_ATTR_SIZE:
+			if (attr->klass == &html_pango_attr_font_size_klass) {
+				HTMLPangoAttrFontSize *size = (HTMLPangoAttrFontSize *) attr;
+				if ((size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != GTK_HTML_FONT_STYLE_SIZE_3 && (size->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != 0)
+					tag = "</FONT>";
+			}
+			break;
 		case PANGO_ATTR_FOREGROUND:
 			tag = "</FONT>";
 			break;
@@ -1552,52 +1639,79 @@ get_length (HTMLObject *self)
 
 /* #define DEBUG_NBSP */
 
+struct TmpDeltaRecord
+{
+	int index;
+	int delta;
+};
+
 static gboolean
-check_last_white (gboolean rv, gint white_space, gunichar last_white, gint *delta_out)
+check_last_white (gint white_space, gunichar last_white, gint *delta_out, gboolean *rv_out)
 {
 	if (white_space > 0 && last_white == ENTITY_NBSP) {
 		(*delta_out) --;
-		rv = TRUE;
+		*rv_out = TRUE;
+
+		return TRUE;
 	}
 
-	return rv;
+	return FALSE;
 }
 
 static gboolean
-check_prev_white (gboolean rv, gint white_space, gunichar last_white, gint *delta_out)
+check_prev_white (gint white_space, gunichar last_white, gint *delta_out, gboolean *rv_out)
 {
 	if (white_space > 0 && last_white == ' ') {
 		(*delta_out) ++;
-		rv = TRUE;
+		*rv_out = TRUE;
+
+		return TRUE;
 	}
 
-	return rv;
+	return FALSE;
+}
+
+static GSList *
+add_change (GSList *list, int index, int delta)
+{
+	struct TmpDeltaRecord *rec = g_new (struct TmpDeltaRecord, 1);
+
+	rec->index = index;
+	rec->delta = delta;
+
+	return g_slist_prepend (list, rec);
 }
 
 static gboolean
-is_convert_nbsp_needed (const gchar *s, gint *delta_out)
+is_convert_nbsp_needed (const gchar *s, gint *delta_out, GSList **changes_out)
 {
 	gunichar uc, last_white = 0;
-	gboolean rv = FALSE;
+	gboolean rv = FALSE, change;
 	gint white_space;
-	const gchar *p, *op;
+	const gchar *p, *op, *bop;
 
 	*delta_out = 0;
 
-	op = p = s;
+	bop = p = s;
 	white_space = 0;
-	while (*p && (uc = g_utf8_get_char (p)) && (p = g_utf8_next_char (p))) {
+	while (*p && (uc = g_utf8_get_char (p)) && (op = p) && (p = g_utf8_next_char (p))) {
 		if (uc == ENTITY_NBSP || uc == ' ') {
-			rv = check_prev_white (rv, white_space, last_white, delta_out);
+			change = check_prev_white (white_space, last_white, delta_out, &rv);
 			white_space ++;
 			last_white = uc;
 		} else {
-			rv = check_last_white (rv, white_space, last_white, delta_out);
+			change = check_last_white (white_space, last_white, delta_out, &rv);
 			white_space = 0;
 		}
-		op = p;
+		if (change)
+			*changes_out = add_change (*changes_out, bop - s, *delta_out);
+		bop = op;
 	}
-	rv = check_last_white (rv, white_space, last_white, delta_out);
+
+	if (check_last_white (white_space, last_white, delta_out, &rv))
+		*changes_out = add_change (*changes_out, op - s, *delta_out);
+
+	*changes_out = g_slist_reverse (*changes_out);
 
 	return rv;
 }
@@ -1605,13 +1719,13 @@ is_convert_nbsp_needed (const gchar *s, gint *delta_out)
 static void
 write_prev_white_space (gint white_space, gchar **fill)
 {
-			if (white_space > 0) {
+	if (white_space > 0) {
 #ifdef DEBUG_NBSP
-				printf ("&nbsp;");
+		printf ("&nbsp;");
 #endif
-				**fill = 0xc2; (*fill) ++;
-				**fill = 0xa0; (*fill) ++;
-			}
+		**fill = 0xc2; (*fill) ++;
+		**fill = 0xa0; (*fill) ++;
+	}
 }
 
 static void
@@ -1625,17 +1739,18 @@ write_last_white_space (gint white_space, gchar **fill)
 	}
 }
 
-static void
-convert_nbsp (gchar *fill, const gchar *p)
+static GSList *
+convert_nbsp (gchar *fill, const gchar *text)
 {
+	GSList *changes = NULL; 
 	gint white_space;
 	gunichar uc;
-	const gchar *op;
+	const gchar *op, *p;
 
 #ifdef DEBUG_NBSP
 	printf ("convert_nbsp: %s --> \"", p);
 #endif
-	op = p;
+	op = p = text;
 	white_space = 0;
 
 	while (*p && (uc = g_utf8_get_char (p)) && (p = g_utf8_next_char (p))) {
@@ -1662,19 +1777,106 @@ convert_nbsp (gchar *fill, const gchar *p)
 #endif
 }
 
+static void
+update_index_interval (int *start_index, int *end_index, GSList *changes)
+{
+	GSList *c;
+	int index, delta;
+
+	index = delta = 0;
+
+	for (c = changes; c && *start_index > index; c = c->next) {
+		struct TmpDeltaRecord *rec = c->data;
+
+		if (*start_index > index && *start_index <= rec->index) {
+			(*start_index) += delta;
+			break;
+		}
+		index = rec->index;
+		delta = rec->delta;
+	}
+
+	if (c == NULL && *start_index > index) {
+		(*start_index) += delta;
+		(*end_index) += delta;
+		return;
+	}
+
+	for (; c && *end_index > index; c = c->next) {
+		struct TmpDeltaRecord *rec = c->data;
+
+		if (*end_index > index && *end_index <= rec->index) {
+			(*start_index) += delta;
+			break;
+		}
+		index = rec->index;
+		delta = rec->delta;
+	}
+
+	if (c == NULL && *end_index > index)
+		(*end_index) += delta;
+}
+
+static gboolean
+update_attributes_filter (PangoAttribute *attr, gpointer data)
+{
+	update_index_interval (&attr->start_index, &attr->end_index, (GSList *) data);
+
+	return FALSE;
+}
+
+static void
+update_attributes (PangoAttrList *attrs, GSList *changes)
+{
+	pango_attr_list_filter (attrs, update_attributes_filter, changes);
+}
+
+static void
+update_links (GSList *links, GSList *changes)
+{
+	GSList *cl;
+
+	for (cl = links; cl; cl = cl->next) {
+		Link *link = (Link *) cl->data;
+		update_index_interval (&link->start_index, &link->end_index, changes);
+	}
+}
+
+static void
+free_changes (GSList *changes)
+{
+	GSList *c;
+
+	for (c = changes; c; c = c->next)
+		g_free (c->data);
+	g_slist_free (changes);
+}
+
 gboolean
 html_text_convert_nbsp (HTMLText *text, gboolean free_text)
 {
+	GSList *changes = NULL;
 	gint delta;
-	gchar *to_free;
 
-	if (is_convert_nbsp_needed (text->text, &delta)) {
-		to_free    = text->text;
+	if (is_convert_nbsp_needed (text->text, &delta, &changes)) {
+		GSList *attrs, *extra_attrs;
+		gchar *to_free;
+
+		to_free = text->text;
 		text->text = g_malloc (strlen (to_free) + delta + 1);
 		text->text_bytes += delta;
 		convert_nbsp (text->text, to_free);
 		if (free_text)
 			g_free (to_free);
+		if (changes) {
+			if (text->attr_list)
+				update_attributes (text->attr_list, changes);
+			if (text->extra_attr_list)
+				update_attributes (text->extra_attr_list, changes);
+			if (text->links)
+				update_links (text->links, changes);
+			free_changes (changes);
+		}
 		html_object_change_set (HTML_OBJECT (text), HTML_CHANGE_ALL);
 		return TRUE;
 	}
@@ -2068,7 +2270,7 @@ html_text_class_init (HTMLTextClass *klass,
 	object_class->split = object_split;
 	object_class->draw = draw;
 	object_class->accepts_cursor = accepts_cursor;
-	object_class->calc_size = calc_size;
+	object_class->calc_size = html_text_real_calc_size;
 	object_class->calc_preferred_width = calc_preferred_width;
 	object_class->calc_min_width = calc_min_width;
 	object_class->fit_line = ht_fit_line;
@@ -2487,42 +2689,48 @@ html_text_append_link (HTMLText *text, gchar *url, gchar *target, gint start_off
 void
 html_text_add_link_full (HTMLText *text, HTMLEngine *e, gchar *url, gchar *target, gint start_index, gint end_index, gint start_offset, gint end_offset)
 {
-	GSList *l, *lnext, *lprev = NULL;
-	Link *link;
+	GSList *l, *lnext, *prev = NULL;
+	Link *link, *plink = NULL;
+
+	cut_links_full (text, start_offset, end_offset, start_index, end_index, 0, 0);
 
 	if (text->links == NULL)
 		html_text_append_link_full (text, url, target, start_index, end_index, start_offset, end_offset);
-	else
-		for (l = text->links; l; l = lnext) {
-			lnext = l->next;
+	else {
+		Link *plink = NULL, *new_link = html_link_new (url, target, start_index, end_index, start_offset, end_offset);
+
+		for (l = text->links; new_link && l; l = l->next) {
 			link = (Link *) l->data;
-			if (link->start_offset >= start_offset && link->end_index <= end_index) {
-				if (lprev)
-					lprev->next = g_slist_delete_link (l, l);
-				else
-					text->links = g_slist_delete_link (l, l);
-				html_link_free (link);
-				continue;
+			if (new_link->start_offset >= link->end_offset) {
+				if (new_link->start_offset == link->end_offset && html_link_equal (link, new_link)) {
+					link->end_offset = end_offset;
+					link->end_index = end_index;
+					html_link_free (new_link);
+					new_link = NULL;
+				} else {
+					l = g_slist_prepend (l, new_link);
+					if (prev)
+						prev->next = l;
+					else
+						text->links = l;
+					link = new_link;
+					new_link = NULL;
+				}
+				if (plink && html_link_equal (plink, link) && plink->start_offset == link->end_offset) {
+					plink->start_offset = link->start_offset;
+					plink->start_index = link->start_index;
+					prev->next = g_slist_remove (prev->next, link);
+					html_link_free (link);
+					link = plink;
+				}
+				plink = link;
+				prev = l;
 			}
-			if (link->start_offset < end_offset && link->end_offset > end_offset)
-				link->start_offset = end_offset;
-			else if (link->end_offset > start_offset && link->end_offset <= end_offset)
-				link->end_offset = start_offset;
-
-			if (link->end_offset <= start_offset) {
-				if (lprev)
-					lprev = g_slist_prepend (l, html_link_new (url, target, start_index, end_index, start_offset, end_offset));
-				else
-					text->links = g_slist_prepend (l, html_link_new (url, target, start_index, end_index, start_offset, end_offset));
-				break;
-			}
-
-			lprev = l;
 		}
 
-	html_text_set_color_in_range (text, html_colorset_get_color (e->settings->color_set, HTMLLinkColor),
-				      start_offset, end_offset);
-	html_text_set_style_in_range (text, GTK_HTML_FONT_STYLE_UNDERLINE, e, start_index, end_index);
+		if (new_link && prev)
+			prev->next = g_slist_prepend (NULL, new_link);
+	}
 }
 
 void
@@ -2765,7 +2973,10 @@ html_pango_attr_font_size_calc (HTMLPangoAttrFontSize *attr, HTMLEngine *e)
 	gint size, base_size, real_size;
 
 	base_size = (attr->style & GTK_HTML_FONT_STYLE_FIXED) ? e->painter->font_manager.fix_size : e->painter->font_manager.var_size;
-	size = (attr->style & GTK_HTML_FONT_STYLE_SIZE_MASK) - GTK_HTML_FONT_STYLE_SIZE_3;
+	if ((attr->style & GTK_HTML_FONT_STYLE_SIZE_MASK) != 0)
+		size = (attr->style & GTK_HTML_FONT_STYLE_SIZE_MASK) - GTK_HTML_FONT_STYLE_SIZE_3;
+	else
+		size = 0;
 	real_size = e->painter->font_manager.magnification * ((gdouble) base_size + (size > 0 ? (1 << size) : size) * base_size/8.0);
 
 	attr->attr_int.value = real_size;
@@ -2885,8 +3096,8 @@ html_text_get_style_conflicts (HTMLText *text, GtkHTMLFontStyle style, gint star
 	return conflicts;
 }
 
-void
-html_text_set_style_in_range (HTMLText *text, GtkHTMLFontStyle style, HTMLEngine *e, gint start_index, gint end_index)
+static void
+html_text_change_attrs (PangoAttrList *attr_list, GtkHTMLFontStyle style, HTMLEngine *e, gint start_index, gint end_index)
 {
 	PangoAttribute *attr;
 
@@ -2895,45 +3106,49 @@ html_text_set_style_in_range (HTMLText *text, GtkHTMLFontStyle style, HTMLEngine
 		attr = pango_attr_weight_new (PANGO_WEIGHT_BOLD);
 		attr->start_index = start_index;
 		attr->end_index = end_index;
-		pango_attr_list_change (text->attr_list, attr);
+		pango_attr_list_change (attr_list, attr);
 	}
 
 	if (style & GTK_HTML_FONT_STYLE_ITALIC) {
 		attr = pango_attr_style_new (PANGO_STYLE_ITALIC);
 		attr->start_index = start_index;
 		attr->end_index = end_index;
-		pango_attr_list_change (text->attr_list, attr);
+		pango_attr_list_change (attr_list, attr);
 	}
 
 	if (style & GTK_HTML_FONT_STYLE_UNDERLINE) {
 		attr = pango_attr_underline_new (PANGO_UNDERLINE_SINGLE);
 		attr->start_index = start_index;
 		attr->end_index = end_index;
-		pango_attr_list_change (text->attr_list, attr);
+		pango_attr_list_change (attr_list, attr);
 	}
 
 	if (style & GTK_HTML_FONT_STYLE_STRIKEOUT) {
 		attr = pango_attr_strikethrough_new (TRUE);
 		attr->start_index = start_index;
 		attr->end_index = end_index;
-		pango_attr_list_change (text->attr_list, attr);
+		pango_attr_list_change (attr_list, attr);
 	}
 
 	if (style & GTK_HTML_FONT_STYLE_FIXED) {
 		attr = pango_attr_family_new (e->painter->font_manager.fixed.face ? e->painter->font_manager.fixed.face : "Monospace");
 		attr->start_index = start_index;
 		attr->end_index = end_index;
-		pango_attr_list_change (text->attr_list, attr);
+		pango_attr_list_change (attr_list, attr);
 	}
 
-	/* size */
-	if (style & GTK_HTML_FONT_STYLE_SIZE_MASK) {
-		attr = html_pango_attr_font_size_new (style);
-		html_pango_attr_font_size_calc ((HTMLPangoAttrFontSize *) attr, e);
-		attr->start_index = start_index;
-		attr->end_index = end_index;
-		pango_attr_list_change (text->attr_list, attr);
-	}
+	/* always apply size so that we can zoom */
+	attr = html_pango_attr_font_size_new (style);
+	html_pango_attr_font_size_calc ((HTMLPangoAttrFontSize *) attr, e);
+	attr->start_index = start_index;
+	attr->end_index = end_index;
+	pango_attr_list_change (attr_list, attr);
+}
+
+void
+html_text_set_style_in_range (HTMLText *text, GtkHTMLFontStyle style, HTMLEngine *e, gint start_index, gint end_index)
+{
+	html_text_change_attrs (text->attr_list, style, e, start_index, end_index);
 }
 
 void
